use crate::{changelog::Changelog, AccountId, Amount, Error, Status, TransactionId};
use serde::{Deserialize, Serialize, Serializer};
use std::ops::Deref;

#[derive(Clone, Debug, Eq, Ord, PartialOrd, PartialEq)]
/// PaymentID
///
/// This is the payment ID. The payment ID has two public members, which is
/// basically the transaction which created this payment and the position in the
/// transaction.
pub struct PaymentId {
    /// Transaction which created this payment
    pub transaction: TransactionId,
    /// This payment position inside the transaction
    pub position: usize,
}

impl PaymentId {
    /// Returns the bytes representation of the PaymentId
    pub fn bytes(&self) -> [u8; 40] {
        let mut bytes = [0u8; 40];
        bytes[0..32].copy_from_slice(self.transaction.deref());
        bytes[32..40].copy_from_slice(&self.position.to_be_bytes());
        bytes
    }
}

impl Serialize for PaymentId {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        let serialized = self.to_string();
        serializer.serialize_str(&serialized)
    }
}

impl ToString for PaymentId {
    fn to_string(&self) -> String {
        format!("{}:{}", self.transaction, self.position)
    }
}

#[derive(Debug, Deserialize, PartialEq, Eq, Serialize, Clone)]
pub struct ChangelogEntry {
    pub reason: String,
    pub status: Status,
    pub spent: Option<SpentInfo>,
}

#[derive(Clone, Debug, Eq, PartialEq, Serialize, Deserialize)]
pub struct SpentInfo {
    pub by: TransactionId,
    pub status: Status,
}

/// Payment
#[derive(Clone, Debug, Eq, PartialEq, Serialize)]
pub struct Payment {
    /// The payment ID
    pub id: PaymentId,
    /// Which account can spend this new payment
    pub to: AccountId,
    /// The amount of the payment
    pub amount: Amount,
    /// The status of the payment. This status is just the status of identical
    /// to the status of the transaction which created this payment.
    pub status: Status,
    #[serde(skip_serializing_if = "Option::is_none")]
    /// What is the spending status of the payment. If the payment is not spent it will be None
    pub spent: Option<SpentInfo>,
    /// The changelog of this payment
    pub changelog: Vec<Changelog<ChangelogEntry>>,
}

impl Payment {
    /// Returns true if the payment is spendable
    #[inline]
    pub fn is_spendable(&self) -> bool {
        self.spent.is_none() && self.status == Status::Settled
    }

    /// Returns true if the payment is spendable or was spent by a given transaction
    pub fn is_spendable_or_was_by(&self, by: &TransactionId) -> bool {
        if let Some(spent) = self.spent.as_ref() {
            if &spent.by != by {
                return false;
            }
        }
        self.status == Status::Settled
    }

    /// Appends a new changelog taking a snapshot of the current object
    pub fn new_changelog(&mut self, reference: &str) {
        let previous_id = if let Some(Ok(previous_id)) = self.changelog.last().map(|x| x.id()) {
            Some(previous_id)
        } else {
            None
        };

        self.changelog.push(Changelog::new(
            previous_id,
            self.id.bytes().to_vec(),
            ChangelogEntry {
                status: self.status.clone(),
                reason: reference.to_owned(),
                spent: self.spent.clone(),
            },
        ));
    }
}
